﻿/**
 * DISCLAIMER: The sample code or utility or tool described herein is provided on an "as is" basis, 
 * without warranty of any kind. UIDAI does not warrant or guarantee the individual success developers 
 * may have in implementing the sample code on their environment. 
 * 
 * UIDAI does not warrant, guarantee or make any representations of any kind with respect to the 
 * sample code and does not make any representations or warranties regarding the use, 
 * results of use, accuracy, timeliness or completeness of any data or information relating 
 * to the sample code. UIDAI disclaims all warranties, express or implied, and in particular, 
 * disclaims all warranties of merchantability, fitness for a particular purpose, and 
 * warranties related to the code, or any service or software related thereto. 
 * 
 * UIDAI is not responsible for and shall not be liable directly or indirectly for any 
 * direct, indirect damages or costs of any type arising out of use or any action 
 * taken by you or others related to the sample code. 
 * 
 * THIS IS NOT A SUPPORTED SOFTWARE.
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Encodings;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.Oiw;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Asn1;
using Org.BouncyCastle.OpenSsl;
using System.IO;
using Decryptor;
using System.Security.Cryptography.X509Certificates;

namespace ECDecryptor
{
    class Decryptor
    {
        private byte[] m_IV;
        private string m_Location = string.Empty;
        private string m_Password = string.Empty;

        public string Password
        {
            set
            {
                this.m_Password = value;
            }
        }
        public string KeyStoreLocation
        {
            set
            {
                this.m_Location = value;

            }
        }
        public byte[] decrypt(byte[] datatosegregate,bool fromUSBToken)
        {
            byte[][] verdata = Utilities.split(datatosegregate, 11);
            byte[] versionNumber = verdata[0];            
            string verNumber = System.Text.Encoding.ASCII.GetString(versionNumber);
            byte[] verData = null;
            if (verNumber.Equals("VERSION_1.0"))
            {
                verData = verdata[1];
            }
            else
            {
                verData = datatosegregate;
            }
            byte[][] data = Utilities.split(verData, 294);
            byte[] publickey = data[0];
            byte[] publicKeyModulus =  GetPublicKeyInfo(publickey);
            byte[] mergedata = data[1];
            byte[][] datafrompad = Utilities.split(mergedata, 32);
            byte[] padding = datafrompad[0];
            byte[] mergedatafromkey = datafrompad[1];
            byte[][] datafromkey = Utilities.split(mergedatafromkey, 256);
            byte[] encryptedsecret = datafromkey[0];
            Console.WriteLine(encryptedsecret.Length);
            byte[] encryptedmessage = datafromkey[1];
            byte[] secreatkey = null;
            if (fromUSBToken)
            {
                // Get AES key from USB Token
                CSPDec dec = new CSPDec();
                secreatkey = dec.Decrypt(encryptedsecret, padding, publicKeyModulus);
                //ECDecryptor.CSPDecryption csp = new ECDecryptor.CSPDecryption();
                //secreatkey = csp.Decrypt(encryptedsecret, padding, publicKeyModulus);
            }
            else
            {
                // Get AES key from private key file
                secreatkey = this.getkey(encryptedsecret, padding);
            }
            byte[] plaintext = this.decryptmessage(encryptedmessage, padding, secreatkey);
            byte[][] result = Utilities.split(plaintext, 32);

            // Console.WriteLine(Encoding.UTF8.GetString(result[1]));
            
            byte[] hashBytes = GenerateHash(result[1]);

            if (Utilities.ParseByteArraytoHexString(hashBytes) != Utilities.ParseByteArraytoHexString(result[0]))
                throw new System.Exception("File is Corrupt or invalid !");

            // Console.WriteLine(Encoding.UTF8.GetString(result[1]));
            return result[1];
        }

        private byte[] GetPublicKeyInfo(byte[] pubkey)
        {
            RsaKeyParameters rsaKeyParam = (RsaKeyParameters)PublicKeyFactory.CreateKey(pubkey);
           // string publicKeyModulus = rsaKeyParam.Modulus.ToString(16);
            return rsaKeyParam.Modulus.ToByteArrayUnsigned();
           // System.Windows.Forms.MessageBox.Show(publicKeyModulus);
        }

        private byte[] GenerateHash(byte[] message)
        {
            System.Security.Cryptography.HashAlgorithm m_hash = new System.Security.Cryptography.SHA256Managed();
            byte[] hashByte = null;
            try
            {
                hashByte = m_hash.ComputeHash(message);
            }
            catch (ArgumentNullException)
            {throw new ArgumentNullException("Invalid Hash");
            }

            catch (ObjectDisposedException)
            {
                throw new ObjectDisposedException("Object is already disposed");
            }

            return hashByte;
        }
        private byte[] decryptmessage(byte[] cmessage, byte[] iVector, byte[] m_Key)
        {
            {
                //// randomly generated number acts as inetialization vector
                m_IV = new byte[16];
                Array.Copy(iVector, 0, m_IV, 0, 16);

                // GenerateAESKey();
                KeyParameter aesKeyParam = ParameterUtilities.CreateKeyParameter("AES", m_Key);
                ParametersWithIV aesIVKeyParam = new ParametersWithIV(aesKeyParam, m_IV);

                IBufferedCipher cipher = CipherUtilities.GetCipher("AES/CFB/NoPadding");
                cipher.Init(false, aesIVKeyParam);
                return cipher.DoFinal(cmessage);
            }

        }

        private byte[] getkey(byte[] encryptkey, byte[] pad)
        {
            //ECDecryptor.CSPDecryption csp = new CSPDecryption();
            //csp.Decrypt(encryptkey, pad);

            IAsymmetricBlockCipher cipher = new OaepEncoding(new RsaEngine(), new Sha256Digest(), pad);
            AsymmetricKeyParameter privateKey = this.GetPrivateKey(this.m_Location, m_Password);
            cipher.Init(false, privateKey);
            return cipher.ProcessBlock(encryptkey, 0, encryptkey.Length);
            //return null;

        }

        ///// <summary>
        ///// Read the Private Key from Key Store and returns RSA Parameters
        ///// </summary>
        ///// <returns>RSA Key Parameters</returns>
        //private RsaKeyParameters GetPrivateKey()
        //{
        //    //IEnumerable<string> privateKey = File.ReadAllLines(this.m_Location);
        //    //IEnumerator<string> privateKeyEnumerator = privateKey.GetEnumerator();

        //    //privateKeyEnumerator.MoveNext();
        //    //string[] modKeyValue = privateKeyEnumerator.Current.Split('=');


        //    //Org.BouncyCastle.Math.BigInteger mod = new Org.BouncyCastle.Math.BigInteger(modKeyValue[1].Trim(), 16);
        //    //privateKeyEnumerator.MoveNext();
        //    //string[] expKeyValue = privateKeyEnumerator.Current.Split('=');
        //    //Org.BouncyCastle.Math.BigInteger exp = new Org.BouncyCastle.Math.BigInteger(expKeyValue[1].Trim(), 16);

        //    //RsaKeyParameters privateRSAKeys = GetPrivateKeyFromPFX(@"C:\UID\and\temp.pfx","qwerty@123");
        //    RsaKeyParameters privateRSAKeys = GetPrivateKeyFromPEM(@"C:\UID\and\tempPEM3.pem", "qwerty@123");
        //    return privateRSAKeys;
        //    //return new RsaKeyParameters(true, mod, exp);

        //}


        private RsaKeyParameters GetPrivateKeyFromPFX(string FileName, string Password)
        {
            FileStream keyFileStream = File.OpenRead(FileName);
            char[] pwd = Password.ToCharArray();
            Org.BouncyCastle.Pkcs.Pkcs12Store store = new Org.BouncyCastle.Pkcs.Pkcs12Store(keyFileStream, pwd);

            System.Collections.IEnumerator h = store.Aliases.GetEnumerator();
            h.MoveNext();
            object o = h.Current;
            string alias = o.ToString();

            Org.BouncyCastle.Pkcs.AsymmetricKeyEntry privAsymKey = store.GetKey(alias);// as RsaKeyParameters;
            RsaKeyParameters privRSAKey = (RsaKeyParameters)privAsymKey.Key;

            bool isPrivate = privRSAKey.IsPrivate;

            if (isPrivate)
            {
                return privRSAKey;
            }
            else
            {
                throw new SystemException("Selected Certificate / PFX - " + FileName + " doesnt have a private key");
            }
        }
        private RsaKeyParameters GetPrivateKeyFromPEM(string FileName, string Password)
        {
            AsymmetricCipherKeyPair keyPair;
            using (var reader = File.OpenText(FileName)) // file containing RSA PKCS1 private key
            {
                PEMPasswordFinder test = new PEMPasswordFinder(Password);
                PemReader temp = new PemReader(reader, new PEMPasswordFinder(Password));
                keyPair = (AsymmetricCipherKeyPair)temp.ReadObject();
                temp.Reader.Close();
            }


            if (keyPair.Private.IsPrivate)
            {
                RsaKeyParameters rsaPriv = (RsaKeyParameters)keyPair.Private;
                return rsaPriv;
            }
            else
            {
                throw new SystemException("Selected PEM file - " + FileName + " does not contains private key");
            }
        }
        private bool IsPrivateKeyPresentinPEMFile()
        {
            return false;
        }
        private bool IsPrivateKeyPresentinPFXFile()
        {
            return false;
        }
        
        public RsaKeyParameters GetPrivateKey(string FileName, string Password)
        {
            char[] delim = {'.'};
            string FileExtension = (FileName.Split(delim))[1];
            bool result = false;
            RsaKeyParameters rsaPrivateKeys = null;

            switch (FileExtension)
            {
                case "pfx":
                case "PFX":
                case "p12":
                case "P12":
                            //result = IsPrivateKeyPresentinPFXFile();
                            //if (result == false)
                            // throw new SystemException("Selected PFX file - " + FileName + " does not contains private key");
                            //else
                        rsaPrivateKeys = GetPrivateKeyFromPFX(FileName, Password);
                    break;
                case "pem":
                case "PEM":
                            //result = IsPrivateKeyPresentinPEMFile();
                            //if (result == false)
                            //    throw new SystemException("Selected PEM file - " + FileName + " does not contains private key");
                            //else
                            //{
                        rsaPrivateKeys = GetPrivateKeyFromPEM(FileName, Password);
                            //}
                    break;
                default:
                    throw new SystemException("Private key should be either in PEM , P12 or PFX format");
            }
            return rsaPrivateKeys;
        }
    }


    public class PEMPasswordFinder : IPasswordFinder
    {
        string m_Password;

        public PEMPasswordFinder(string password)
        {
            m_Password = password;
        }

        #region IPasswordFinder Members
        public char[] GetPassword()
        {
            return m_Password.ToCharArray();   
        }
        #endregion
    }
}

